#include <string.h>
#include <assert.h>
#include <stdio.h>
#include <ctype.h>
#include "../berLame_src/Tga.h"
#include "../berLame_src/Jpeg.h"

#define for if(0) {} else for

const char *p_s_32bit_decompress_routine =
	"unsigned __int32 *p_DecompressRLEImage(const unsigned __int32 *p_image, int n_size)\n"
	"{\n"
	"	unsigned __int32 *p_buffer;\n"
	"	if(!(p_buffer = new unsigned __int32[n_size]))\n"
	"		return 0;\n"
	"\n"
	"	const unsigned __int32 *p_end = p_buffer + n_size;\n"
	"	for(unsigned __int32 *p_dest = p_buffer; p_dest != p_end;) {\n"
	"		bool b_compress = (*p_image >> 31) != 0;\n"
	"		int n_run_length = *p_image & ((1u << 31) - 1);\n"
	"\n"
	"		++ p_image; // skip run length\n"
	"\n"
	"		if(b_compress) {\n"
	"			unsigned __int32 n_value = *p_image;\n"
	"			++ p_image;\n"
	"			// get color\n"
	"\n"
	"			for(unsigned __int32 *p_end2 = p_dest + n_run_length; p_dest != p_end2; ++ p_dest)\n"
	"				*p_dest = n_value;\n"
	"			// fill\n"
	"		} else {\n"
	"			for(unsigned __int32 *p_end2 = p_dest + n_run_length;\n"
	"			   p_dest != p_end2; ++ p_dest, ++ p_image)\n"
	"				*p_dest = *p_image;\n"
	"			// copy data\n"
	"		}\n"
	"	}\n"
	"	// decompress\n"
	"\n"
	"	return p_buffer;\n"
	"};\n";

const char *p_s_8bit_decompress_routine =
	"unsigned char *p_DecompressRLEImage(const unsigned char *p_image, int n_size)\n"
	"{\n"
	"	unsigned char *p_buffer;\n"
	"	if(!(p_buffer = new unsigned char[n_size]))\n"
	"		return 0;\n"
	"\n"
	"	const unsigned char *p_end = p_buffer + n_size;\n"
	"	for(unsigned char *p_dest = p_buffer; p_dest != p_end;) {\n"
	"		bool b_compress = (*p_image >> 7) != 0;\n"
	"		int n_run_length = *p_image & ((1u << 7) - 1);\n"
	"\n"
	"		++ p_image; // skip run length\n"
	"\n"
	"		if(b_compress) {\n"
	"			unsigned char n_value = *p_image;\n"
	"			++ p_image;\n"
	"			// get color\n"
	"\n"
	"			for(unsigned char *p_end2 = p_dest + n_run_length; p_dest != p_end2; ++ p_dest)\n"
	"				*p_dest = n_value;\n"
	"			// fill\n"
	"		} else {\n"
	"			for(unsigned char *p_end2 = p_dest + n_run_length;\n"
	"			   p_dest != p_end2; ++ p_dest, ++ p_image)\n"
	"				*p_dest = *p_image;\n"
	"			// copy data\n"
	"		}\n"
	"	}\n"
	"	// decompress\n"
	"\n"
	"	return p_buffer;\n"
	"};\n";

bool Write_Header(const TBmp *p_image, bool b_gray, bool b_use_rle,
	bool b_rle_decoder, bool b_verbose, const char *p_s_infile, const char *p_s_outfile)
{
	FILE *p_fw;
	if(!(p_fw = fopen(p_s_outfile, "w")))
		return false;

	do {
		char p_s_header_name[256] = "__";
		strncpy(p_s_header_name + 2, p_s_outfile, 240);
		p_s_header_name[240] = 0;
		for(char *p_cur = p_s_header_name + 2, *p_end = p_s_header_name +
		   strlen(p_s_header_name); p_cur != p_end; ++ p_cur) {
			if(isalpha(*p_cur))
				*p_cur = toupper(*p_cur);
			else if(!isdigit(*p_cur))
				*p_cur = '_';
		}
		strcat(p_s_header_name, "_INCLUDED");
		// create header name

		char p_s_variable_name[256] = "";
		const char *p_src = p_s_infile;
		int n_length = strlen(p_src);
		if(n_length > sizeof(p_s_variable_name) / sizeof(p_s_variable_name[0]))
			n_length = sizeof(p_s_variable_name) / sizeof(p_s_variable_name[0]);
		for(char *p_cur = p_s_variable_name, *p_end = p_s_variable_name +
		   n_length; p_cur != p_end; ++ p_cur, ++ p_src) {
			if(isalpha(*p_src))
				*p_cur = tolower(*p_src);
			else if(!isdigit(*p_src)) {
				if(*(p_cur - 1) != '_')
					*p_cur = '_';
			} else
				*p_cur = *p_src;
		}
		// create variable name

		if(!fprintf(p_fw, "/*\n *\timage header file\n *\tcreated from \'%s\'\n */\n\n",
		   p_s_infile))
			break;
		// print starting comment

		if(!fprintf(p_fw, "#ifndef %s\n#define %s\n\n", p_s_header_name, p_s_header_name))
			break;
		// print include-guard

		if(!fprintf(p_fw, "const unsigned %s *p_%s[] = {\n\t", (b_gray)?
		   "char" : "__int32", p_s_variable_name))
			break;
		// print data section begin

		int n_records_line = 80 / (((b_gray)? 4 : 10) + 2) + 1;
		if(!b_use_rle) {
			int n_count = n_records_line;
			for(unsigned __int32 *p_pixel = p_image->p_buffer,
			   *p_end = p_image->p_buffer + (p_image->n_width * p_image->n_height);
			   p_pixel != p_end; ++ p_pixel) {
				fprintf(p_fw, (b_gray)? "0x%02x%s" : "0x%08x%s", *p_pixel,
					(-- n_count && p_pixel + 1 < p_end)? ", " : ((p_pixel + 1 < p_end)? ",\n\t" : "\n"));
				if(!n_count)
					n_count = n_records_line;
			}
			// print uncompressed data
		} else {
			int n_run_bit = (b_gray)? 1 << 7 : 1 << 31;
			int n_max_repeats = ((b_gray)? 1 << 7 : 1 << 31) - 1;

			int n_count = n_records_line;
			for(unsigned __int32 *p_pixel = p_image->p_buffer,
			   *p_end = p_image->p_buffer + (p_image->n_width * p_image->n_height);
			   p_pixel != p_end; ++ p_pixel) {
				int n_remains = p_end - p_pixel;
				if(n_remains > n_max_repeats)
					n_remains = n_max_repeats;
				bool b_compress = n_remains > 2 && *p_pixel == *(p_pixel + 1) &&
					*p_pixel == *(p_pixel + 2);
				int n_run_length = n_remains;
				for(int i = 1; i < n_remains; ++ i) {
					if(b_compress && *p_pixel != *(p_pixel + i)) {
						n_run_length = i;
						break;
					} else if(!b_compress && i + 1 < n_remains &&
					   *(p_pixel + i - 1) == *(p_pixel + i) &&
					   *(p_pixel + i - 1) == *(p_pixel + i + 1)) {
						n_run_length = i;
						break;
					}
				}
				// determine run length and wheter to compress data

				fprintf(p_fw, (b_gray)? "0x%02x%s" : "0x%08x%s", n_run_length |
					((b_compress)? n_run_bit : 0), (-- n_count && p_pixel + 1 < p_end)? ", " : ((p_pixel + 1 < p_end)? ",\n\t" : "\n"));
				if(!n_count)
					n_count = n_records_line;
				// print run length

				if(b_compress) {
					p_pixel += n_run_length - 1;
					fprintf(p_fw, (b_gray)? "0x%02x%s" : "0x%08x%s", *p_pixel,
						(-- n_count && p_pixel + 1 < p_end)? ", " : ((p_pixel + 1 < p_end)? ",\n\t" : "\n"));
					if(!n_count)
						n_count = n_records_line;
					// print compressed pixels
				} else {
					for(unsigned __int32 *p_end2 = p_pixel + n_run_length;
					   p_pixel != p_end2; ++ p_pixel) {
						fprintf(p_fw, (b_gray)? "0x%02x%s" : "0x%08x%s", *p_pixel,
							(-- n_count && p_pixel + 1 < p_end)? ", " : ((p_pixel + 1 < p_end)? ",\n\t" : "\n"));
						if(!n_count)
							n_count = n_records_line;
					}
					-- p_pixel; // ++ in the for command
					// print uncompressed pixels
				}

				assert(p_pixel < p_end);
			}
			// print image data
		}
		// print data

		if(!fprintf(p_fw, "};\n\n"))
			break;
		// print data section end

		if(!fprintf(p_fw, "const unsigned int n_%s_width = %d;\n"
		   "const unsigned int n_%s_height = %d;\n", p_s_variable_name,
		   p_image->n_width, p_s_variable_name, p_image->n_height))
			break;
		// print image dimensions

		if(b_rle_decoder) {
			if(!fprintf(p_fw, "\n%s", (b_gray)? p_s_8bit_decompress_routine :
			   p_s_32bit_decompress_routine))
				break;
		}
		// include rle decoder source

		if(!fprintf(p_fw, "\n#endif //%s\n", p_s_header_name))
			break;
		// print include-guard end

		fclose(p_fw);
		return true;
	} while(0);

	fclose(p_fw);
	return false;
}

int main(int n_arg_num, const char **p_arg_list)
{
	if(n_arg_num < 3) {
		fprintf(stderr, "error: missing program arguments.\n"
			"usage:\n"
			"ImageToHeader <image-file> <header-file> [--verbose] [--rle]\n"
			"\t[--rle-decoder] [--rgb|--gray]\n"
			"where image-file is source file, containing image in jpeg or tga format,\n"
			"and header-file is output file to contain image.\n"
			"--verbose parameter turns on verbose output\n"
			"--rle turns on run-length compression (header is smaller)\n"
			"--rle-decoder includes decoding function in header\n"
			"--rgb forces RGB(A) colors\n"
			"--gray forces grayscale colors (single channel)\n");
		return 0;
	}

	bool b_verbose = false;
	bool b_use_rle = false;
	bool b_rle_decoder = false;
	bool b_force_gray = false;

	const char *p_s_infile = p_arg_list[1];
	const char *p_s_outfile = p_arg_list[2];

	for(int i = 3; i < n_arg_num; ++ i) {
		if(!strcmp(p_arg_list[i], "--verbose"))
			b_verbose = true;
		else if(!strcmp(p_arg_list[i], "--rle"))
			b_use_rle = true;
		else if(!strcmp(p_arg_list[i], "--rle-decoder"))
			b_rle_decoder = true;
		else if(!strcmp(p_arg_list[i], "--gray"))
			b_force_gray = true;
		else if(!strcmp(p_arg_list[i], "--rgb"))
			b_force_gray = false;
		else
			fprintf(stderr, "warning: unknown parameter: \'%s\' (skipping)\n", p_arg_list[i]);
	}
	// parse command-line

	TBmp *p_image;
	if(strrchr(p_s_infile, '.') && !_stricmp(strrchr(p_s_infile, '.'), ".tga")) {
		if(b_verbose)
			printf("loading image in TGA format ...\n");
		p_image = p_Load_TGA(p_s_infile);
	} else if(strrchr(p_s_infile, '.') && (!_stricmp(strrchr(p_s_infile, '.'), ".jpg") ||
	   !_stricmp(strrchr(p_s_infile, '.'), ".jpeg"))) {
		if(b_verbose)
			printf("loading image in JPEG format ...\n");
		FILE *p_fr;
		if(!(p_fr = fopen(p_s_infile, "rb")))
			p_image = 0;
		else {
			CJpegDecoder decoder;
			p_image = decoder.p_Decode_JPEG(p_fr);
			fclose(p_fr);
		}
	} else {
		if(b_verbose)
			printf("attempting to load image in TGA format ...\n");
		p_image = p_Load_TGA(p_s_infile);
		if(!p_image) {
			if(b_verbose)
				printf("attempting to load image in JPEG format ...\n");
			FILE *p_fr;
			if(!(p_fr = fopen(p_s_infile, "rb")))
				p_image = 0;
			else {
				CJpegDecoder decoder;
				p_image = decoder.p_Decode_JPEG(p_fr);
				fclose(p_fr);
			}
		}
	}
	if(!p_image) {
		fprintf(stderr, "error: error loading image file \'%s\'\n", p_s_infile);
		return -1;
	}
	// load image (try to detect format)

	if(b_force_gray) {
		for(unsigned __int32 *p_pixel = p_image->p_buffer,
		   *p_end = p_image->p_buffer + (p_image->n_width * p_image->n_height);
		   p_pixel != p_end; ++ p_pixel) {
			unsigned __int32 n_value = *p_pixel;
			n_value = (unsigned __int32)(0.299f * ((n_value >> 16) & 0xff) +
				0.587f * ((n_value >> 8) & 0xff) + 0.114f * (n_value & 0xff));
			*p_pixel = (n_value <= 255)? n_value : 255;
		}
	}
	// convert to grayscale

	if(b_verbose)
		printf("writing image data ...");

	if(!Write_Header(p_image, b_force_gray, b_use_rle,
	   b_rle_decoder, b_verbose, p_s_infile, p_s_outfile)) {
		fprintf(stderr, "\nerror: error writing header file \'%s\'\n", p_s_outfile);
		delete[] p_image->p_buffer;
		delete p_image;
		return -1;
	}

	if(b_verbose)
		printf(" ok\n");

	delete[] p_image->p_buffer;
	delete p_image;

	return 0;
}
